/*gpgmepp_version.h - Version macros for gpgmepp
  Copyright (C) 2016, Intevation GmbH

  This file is part of GPGME++.

  GPGME++ is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  GPGME++ is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Library General Public License for more details.

  You should have received a copy of the GNU Library General Public License
  along with GPGME++; see the file COPYING.LIB.  If not, write to the
  Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
  Boston, MA 02110-1301, USA.
*/


#ifndef GPGMEPP_VERSION_H
#define GPGMEPP_VERSION_H

#define GPGMEPP_VERSION_STRING "1.23.2"
#define GPGMEPP_VERSION_MAJOR 1
#define GPGMEPP_VERSION_MINOR 23
#define GPGMEPP_VERSION_PATCH 2
#define GPGMEPP_VERSION ((1<<16)|(23<<8)|(2))
#endif
