/* cJSON.h
 * Copyright (c) 2009 Dave Gamble
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *
 * SPDX-License-Identifier: MIT
 *
 * Note that this code has been modified from the original code taken
 * from cjson-code-58.zip.
 */

#ifndef cJSON_h
#define cJSON_h

#ifdef __cplusplus
extern "C"
{
#if 0 /*(to make Emacs auto-indent happy)*/
}
#endif
#endif

/* cJSON Types: */
#define cJSON_False  0
#define cJSON_True   1
#define cJSON_NULL   2
#define cJSON_Number 3
#define cJSON_String 4
#define cJSON_Array  5
#define cJSON_Object 6

#define cJSON_IsReference 256

/* The cJSON structure: */
typedef struct cJSON
{
  /* next/prev allow you to walk array/object chains. Alternatively,
     use GetArraySize/GetArrayItem/GetObjectItem */
  struct cJSON *next, *prev;

  /* An array or object item will have a child pointer pointing to a
     chain of the items in the array/object. */
  struct cJSON *child;

  int type;		/* The type of the item, as above. */

  char *valuestring;	/* The item's string, if type==cJSON_String */
  int valueint;		/* The item's number, if type==cJSON_Number */
  double valuedouble;	/* The item's number, if type==cJSON_Number */

  /* The item's name string, if this item is the child of, or is in
     the list of subitems of an object. */
  char *string;
} cJSON;

typedef struct cJSON *cjson_t;

/* Macros to test the type of an object.  */
#define cjson_is_boolean(a) (!((a)->type & ~1))
#define cjson_is_false(a)   ((a)->type == cJSON_False)
#define cjson_is_true(a)    ((a)->type == cJSON_True)
#define cjson_is_null(a)    ((a)->type == cJSON_NULL)
#define cjson_is_number(a)  ((a)->type == cJSON_Number)
#define cjson_is_string(a)  ((a)->type == cJSON_String)
#define cjson_is_array(a)   ((a)->type == cJSON_Array)
#define cjson_is_object(a)  ((a)->type == cJSON_Object)

/* Supply a block of JSON, and this returns a cJSON object you can
   interrogate. Call cJSON_Delete when finished. */
extern cJSON *cJSON_Parse(const char *value, size_t *r_erroff);

/* Render a cJSON entity to text for transfer/storage. Free the char*
   when finished. */
extern char  *cJSON_Print(cJSON *item);

/* Render a cJSON entity to text for transfer/storage without any
   formatting. Free the char* when finished. */
extern char  *cJSON_PrintUnformatted(cJSON *item);

/* Delete a cJSON entity and all subentities. */
extern void   cJSON_Delete(cJSON *c);

/* Returns the number of items in an array (or object). */
extern int    cJSON_GetArraySize(cJSON *array);

/* Retrieve item number "item" from array "array". Returns NULL if
   unsuccessful. */
extern cJSON *cJSON_GetArrayItem(cJSON *array,int item);

/* Get item "string" from object. Case insensitive. */
extern cJSON *cJSON_GetObjectItem(cJSON *object,const char *string);

/* These calls create a cJSON item of the appropriate type. */
extern cJSON *cJSON_CreateNull(void);
extern cJSON *cJSON_CreateTrue(void);
extern cJSON *cJSON_CreateFalse(void);
extern cJSON *cJSON_CreateBool(int b);
extern cJSON *cJSON_CreateNumber(double num);
extern cJSON *cJSON_CreateString(const char *string);
extern cJSON *cJSON_CreateStringConvey (char *string);
extern cJSON *cJSON_CreateArray(void);
extern cJSON *cJSON_CreateObject(void);

/* These utilities create an Array of count items. */
extern cJSON *cJSON_CreateIntArray(const int *numbers,int count);
extern cJSON *cJSON_CreateFloatArray(const float *numbers,int count);
extern cJSON *cJSON_CreateDoubleArray(const double *numbers,int count);
extern cJSON *cJSON_CreateStringArray(const char **strings,int count);

/* Append item to the specified array. */
extern void cJSON_AddItemToArray(cJSON *array, cJSON *item);

/* Append item to the specified object. */
extern cJSON *cJSON_AddItemToObject(cJSON *object, const char *name,
                                    cJSON *item);
extern cJSON *cJSON_AddNullToObject (cJSON *object, const char *name);
extern cJSON *cJSON_AddTrueToObject (cJSON *object, const char *name);
extern cJSON *cJSON_AddFalseToObject (cJSON *object, const char *name);
extern cJSON *cJSON_AddBoolToObject (cJSON *object, const char *name, int b);
extern cJSON *cJSON_AddNumberToObject (cJSON *object, const char *name,
                                       double num);
extern cJSON *cJSON_AddStringToObject (cJSON *object, const char *name,
                                       const char *string);

/* Append reference to item to the specified array/object. Use this
   when you want to add an existing cJSON to a new cJSON, but don't
   want to corrupt your existing cJSON. */
extern void cJSON_AddItemReferenceToArray(cJSON *array, cJSON *item);
extern void cJSON_AddItemReferenceToObject(cJSON *object,
                                           const char *string,cJSON *item);

/* Remove/Detach items from Arrays/Objects. */
extern cJSON *cJSON_DetachItemFromArray(cJSON *array,int which);
extern void   cJSON_DeleteItemFromArray(cJSON *array,int which);
extern cJSON *cJSON_DetachItemFromObject(cJSON *object,const char *string);
extern void   cJSON_DeleteItemFromObject(cJSON *object,const char *string);

/* Update array items. */
extern void cJSON_ReplaceItemInArray(cJSON *array,int which,cJSON *newitem);
extern void cJSON_ReplaceItemInObject(cJSON *object,
                                      const char *string, cJSON *newitem);

/* Duplicate a cJSON item */
extern cJSON *cJSON_Duplicate(cJSON *item,int recurse);

/* Duplicate will create a new, identical cJSON item to the one you
   pass, in new memory that will need to be released. With recurse!=0,
   it will duplicate any children connected to the item.  The
   item->next and ->prev pointers are always zero on return from
   Duplicate. */

/* ParseWithOpts allows you to require (and check) that the JSON is
   null terminated, and to retrieve the pointer to the final byte
   parsed. */
extern cJSON *cJSON_ParseWithOpts(const char *value,
                                  const char **return_parse_end,
                                  int require_null_terminated,
                                  size_t *r_erroff);

extern void cJSON_Minify(char *json);

/* When assigning an integer value, it needs to be propagated to
   valuedouble too. */
#define cJSON_SetIntValue(object,val) \
  ((object)?(object)->valueint=(object)->valuedouble=(val):(val))

#ifdef __cplusplus
}
#endif

#endif /* cJSON_h */
